@@ -1,7 +1,7 @@
 __all__ = [
     "EncodeError", "ParseError",
     "Asn1Data", "Class", "Structure", "Identifier",
-    "decode_identifier", "decode", "encode",
+    "decode", "encode",
 ]
 
 from enum import IntEnum
@@ -12,14 +12,17 @@ from snmp.utils import subbytes
 
 Asn1Data = Union[bytes, subbytes]
 
+@final
 class EncodeError(SNMPException):
     """Failure to encode a payload under ASN.1 Basic Encoding Rules."""
     pass
 
+@final
 class ParseError(IncomingMessageError):
     """Failure to translate a byte string into an ASN.1 object."""
     pass
 
+@final
 class Class(IntEnum):
     """Named constants for the class bits of an ASN.1 BER identifier."""
     UNIVERSAL         = 0
@@ -27,77 +30,82 @@ class Class(IntEnum):
     CONTEXT_SPECIFIC  = 2
     PRIVATE           = 3
 
+@final
 class Structure(IntEnum):
     """Named constants for the constructed bit of an ASN.1 BER identifier."""
     PRIMITIVE     = 0
     CONSTRUCTED   = 1
 
+TIdentifier = TypeVar("TIdentifier", bound="Identifier")
+
+@final
 class Identifier(NamedTuple):
     """Represents an ASN.1 BER identifier."""
     cls: Class
     structure: Structure
     tag: int
 
-def decode_identifier(data: subbytes) -> Identifier:
-    """Extract the identifier from an ASN.1 BER string.
+    @classmethod
+    def decode(cls: Type[TIdentifier], data: subbytes) -> TIdentifier:
+        """Extract the identifier from an ASN.1 BER string.
 
-    This function decodes the identifier portion of a BER string an returns
-    it as an :class:`Identifier` object. As a side-effect, it advances the
-    start of the `data` argument to point to the byte immediately after the
-    identifier.
-    """
-    try:
-        byte = data.consume()
-    except IndexError as err:
-        raise ParseError("Missing identifier") from err
+        This function decodes the identifier portion of a BER string an returns
+        it as an :class:`Identifier` object. As a side-effect, it advances the
+        start of the `data` argument to point to the byte immediately after the
+        identifier.
+        """
+        try:
+            byte = data.consume()
+        except IndexError as err:
+            raise ParseError("Missing identifier") from err
 
-    cls         = (byte & 0xc0) >> 6
-    structure   = (byte & 0x20) >> 5
-    tag         = (byte & 0x1f)
+        class_      = (byte & 0xc0) >> 6
+        structure   = (byte & 0x20) >> 5
+        tag         = (byte & 0x1f)
 
-    if tag == 0x1f:
-        tag = 0
-        byte = 0x80
-        while byte & 0x80:
-            try:
-                byte = data.consume()
-            except IndexError as err:
-                raise ParseError("Incomplete identifier") from err
+        if tag == 0x1f:
+            tag = 0
+            byte = 0x80
+            while byte & 0x80:
+                try:
+                    byte = data.consume()
+                except IndexError as err:
+                    raise ParseError("Incomplete identifier") from err
 
-            tag <<= 7
-            tag |= byte & 0x7f
+                tag <<= 7
+                tag |= byte & 0x7f
 
-    return Identifier(Class(cls), Structure(structure), tag)
+        return cls(Class(class_), Structure(structure), tag)
 
-def encode_identifier(i: Identifier) -> bytes:
-    """Encode an Identifier under ASN.1 Basic Encoding Rules."""
-    byte = (
-        ((i.cls       << 6) & 0xc0) |
-        ((i.structure << 5) & 0x20)
-    )
+    def encode(self) -> bytes:
+        """Encode an Identifier under ASN.1 Basic Encoding Rules."""
+        byte = (
+            ((self.cls       << 6) & 0xc0) |
+            ((self.structure << 5) & 0x20)
+        )
 
-    arr = bytearray()
-    if i.tag < 0x1f:
-        byte |= i.tag
-    else:
-        byte |= 0x1f
+        arr = bytearray()
+        if self.tag < 0x1f:
+            byte |= self.tag
+        else:
+            byte |= 0x1f
 
-        tag = i.tag
-        indicator = 0
-        while tag:
-            arr.append(indicator | (tag & 0x7f))
-            indicator = 0x80
-            tag >>= 7
+            tag = self.tag
+            indicator = 0
+            while tag:
+                arr.append(indicator | (tag & 0x7f))
+                indicator = 0x80
+                tag >>= 7
 
-    arr.append(byte)
-    return bytes(reversed(arr))
+        arr.append(byte)
+        return bytes(reversed(arr))
 
 def decode_length(data: subbytes) -> int:
     """Decode the length field of an ASN.1 BER string.
 
     The provided `data` argument should contain most of a BER string,
     starting after the identifier. The most natural way to use it is to pass
-    the same object first to :func:`decode_identifier` and then to
+    the same object first to :func:`Identifier.decode` and then to
     :func:`decode_length`. This function will decode the length field and
     return it as an :class:`int`, modifying the `data` argument to start
     immediately after the length field.
@@ -242,7 +250,7 @@ def decode(
     preserve a reference to the complete message.
     """
     data = subbytes(data)
-    identifier = decode_identifier(data)
+    identifier = Identifier.decode(data)
 
     if expected is not None and identifier != expected:
         raise ParseError("Identifier does not match expected type")
@@ -270,4 +278,4 @@ def decode(
 
 def encode(identifier: Identifier, data: bytes) -> bytes:
     """Encode a message under ASN.1 Basic Encoding Rules."""
-    return encode_identifier(identifier) + encode_length(len(data)) + data
+    return identifier.encode() + encode_length(len(data)) + data
