@@ -12,11 +12,13 @@ from snmp.types import SEQUENCE, Integer
 from snmp.utils import typename
 
 class Dispatcher(TransportListener):
-    def __init__(self):
+    def __init__(self, multiplexor):
         self.lock = threading.Lock()
         self.msgProcessors = {}
-        self.threads = {}
+
         self.transports = {}
+        self.multiplexor = multiplexor
+        self.thread = None
 
     def addMessageProcessor(self, mp):
         with self.lock:
@@ -24,28 +26,30 @@ class Dispatcher(TransportListener):
 
     def connectTransport(self, transport):
         domain = transport.DOMAIN
+        if domain in self.transports:
+            if self.transports.get(domain) is transport:
+                return
 
-        with self.lock:
-            if domain in self.threads:
-                if self.transports.get(domain) is transport:
-                    return
+            errmsg = "This {} instance is already connected to {}"
+            raise ValueError(errmsg.format(typename(self), str(domain)))
 
-                errmsg = "This {} instance is already connected to {}"
-                raise ValueError(errmsg.format(typename(self), str(domain)))
+        if self.thread is not None:
+            self.multiplexor.stop()
+            self.thread.join()
 
-            thread = threading.Thread(
-                target=transport.listen,
-                args=(self,))
+        self.multiplexor.register(transport)
+        self.thread = threading.Thread(
+            target=self.multiplexor.listen,
+            args=(self,)
+        )
 
-            thread.start()
-
-            self.transports[domain] = transport
-            self.threads[domain] = thread
+        self.thread.start()
+        self.transports[domain] = transport
 
     def hear(self, transport, address, data):
         try:
             try:
-                msgVersion, message = MessageBase.decodeVersion(data)
+                msgVersion = MessageVersion.decode(data).version
             except ParseError:
                 return
 
@@ -56,7 +60,7 @@ class Dispatcher(TransportListener):
                     return
 
             try:
-                message, handle = mp.prepareDataElements(message)
+                message, handle = mp.prepareDataElements(data)
                 handle.push(message)
             except IncomingMessageError:
                 return
@@ -67,13 +71,13 @@ class Dispatcher(TransportListener):
             pass
 
     def sendPdu(self, locator, mpm, pdu, handle, *args, **kwargs):
-        with self.lock:
-            try:
-                transport = self.transports[locator.domain]
-            except KeyError as err:
-                domain = str(TransportDomain(locator.domain))
-                raise ValueError("{} is not enabled".format(domain)) from err
+        try:
+            transport = self.transports[locator.domain]
+        except KeyError as err:
+            domain = str(TransportDomain(locator.domain))
+            raise ValueError("{} is not enabled".format(domain)) from err
 
+        with self.lock:
             try:
                 mp = self.msgProcessors[mpm]
             except KeyError as err:
@@ -84,15 +88,10 @@ class Dispatcher(TransportListener):
         transport.send(locator.address, msg)
 
     def shutdown(self):
-        with self.lock:
-            for transport in self.transports.values():
-                transport.stop()
+        if self.thread is not None:
+            self.multiplexor.stop()
+            self.thread.join()
 
-            for thread in self.threads.values():
-                thread.join()
-
-            for transport in self.transports.values():
-                transport.close()
-
-            self.transports.clear()
-            self.threads.clear()
+        self.multiplexor.close()
+        self.transports.clear()
+        self.thread = None
