@@ -14,13 +14,14 @@ from snmp.transport.udp import *
 class Engine:
     TRANSPORTS = {
         cls.DOMAIN: cls for cls in [
-            UdpTransport,
+            UdpIPv4Socket,
+            UdpIPv6Socket,
         ]
     }
 
     def __init__(self,
         defaultVersion=MessageProcessingModel.SNMPv3,
-        defaultDomain=TransportDomain.UDP,
+        defaultDomain=TransportDomain.UDP_IPv4,
         defaultSecurityModel=SecurityModel.USM,
         defaultCommunity=b"",
         autowait=True
@@ -32,7 +33,7 @@ class Engine:
         self.defaultCommunity       = defaultCommunity
         self.autowaitDefault        = autowait
 
-        self.dispatcher = Dispatcher()
+        self.dispatcher = Dispatcher(UdpMultiplexor())
         self.transports = set()
 
         self.mpv1 = None
@@ -43,7 +44,7 @@ class Engine:
     @property
     def usm(self):
         if self._usm is None:
-            self._usm = UsmAdmin()
+            self._usm = UserBasedSecurityModule()
         return self._usm
 
     def __enter__(self):
@@ -120,7 +121,7 @@ class Engine:
                     namespace,
                 )
 
-            self.mpv3.addSecurityModuleIfNeeded(self.usm.securityModule)
+            self.mpv3.addSecurityModuleIfNeeded(self.usm)
 
             return SNMPv3UsmManager(
                 self.dispatcher,
