@@ -5,9 +5,12 @@ __all__ = [
 from abc import abstractmethod
 import enum
 
+from snmp.types import *
 from snmp.typing import *
 from snmp.utils import *
 
+TMessage = TypeVar("TMessage", bound="Sequence")
+
 class SecurityLevel:
     def __init__(self, auth: Any = False, priv: Any = False) -> None:
         a = bool(auth)
@@ -72,23 +75,20 @@ class SecurityParameters:
             self.securityName,
         )
 
-class SecurityModule:
+class SecurityModule(Generic[TMessage]):
     MODEL: ClassVar[SecurityModel]
 
     @abstractmethod
     def processIncoming(self,
-        msg: subbytes,
-        securityLevel: SecurityLevel,
+        message: TMessage,
         timestamp: Optional[float] = None,
-    ) -> Tuple[SecurityParameters, bytes]:
+    ) -> SecurityParameters:
         ...
 
     @abstractmethod
     def prepareOutgoing(self,
-        header: bytes,
-        data: bytes,
+        message: TMessage,
         engineID: bytes,
         securityName: bytes,
-        securityLevel: SecurityLevel,
     ) -> bytes:
         ...
