@@ -1,17 +1,18 @@
 __all__ = [
     "InvalidEngineID", "InvalidUserName", "InvalidSecurityLevel",
-    "UserBasedSecurityModule", "UsmAdmin",
+    "AuthProtocol", "PrivProtocol", "UserBasedSecurityModule",
 ]
 
 from abc import abstractmethod
 import threading
 
 from time import time
-from snmp.ber import decode, encode
+from snmp.ber import *
 from snmp.exception import IncomingMessageError
-from snmp.types import *
+from snmp.message.v3 import *
 from snmp.security import *
 from snmp.security.levels import *
+from snmp.types import *
 from snmp.typing import *
 from snmp.utils import *
 
@@ -22,9 +23,18 @@ class AuthProtocol:
 
     @classmethod
     @abstractmethod
-    def localize(cls, secret: bytes, engineID: bytes) -> bytes:
+    def computeKey(cls, secret: bytes) -> bytes:
         ...
 
+    @classmethod
+    @abstractmethod
+    def localizeKey(cls, key: bytes, engineID: bytes) -> bytes:
+        ...
+
+    @classmethod
+    def localize(cls, secret: bytes, engineID: bytes) -> bytes:
+        return cls.localizeKey(cls.computeKey(secret), engineID)
+
     @property
     @abstractmethod
     def msgAuthenticationParameters(self) -> bytes:
@@ -149,7 +159,12 @@ class TimeKeeper:
                 self.table[engineID] = entry
 
             withinTimeWindow = False
-            if auth or not entry.authenticated:
+            if not entry.authenticated:
+                entry.snmpEngineBoots = msgBoots
+                entry.latestBootTime = timestamp - msgTime
+                entry.latestReceivedEngineTime = msgTime
+                withinTimeWindow = True
+            elif auth:
                 if msgBoots > entry.snmpEngineBoots:
                     entry.snmpEngineBoots = msgBoots
                     entry.latestBootTime = timestamp
@@ -178,6 +193,32 @@ class TimeKeeper:
             return withinTimeWindow
 
 class Credentials:
+    def __init__(self,
+        authProtocol: Optional[Type[AuthProtocol]] = None,
+        authSecret: Optional[bytes] = None,
+        privProtocol: Optional[PrivProtocol] = None,
+        privSecret: Optional[bytes] = None,
+        secret: bytes = b"",
+    ) -> None:
+        self.authProtocol = None
+        self.authKey = None
+        self.privProtocol = None
+        self.privKey = None
+
+        if authProtocol is None:
+            self.maxSecurityLevel = noAuthNoPriv
+        else:
+            if privProtocol is None:
+                self.maxSecurityLevel = authNoPriv
+            else:
+                self.maxSecurityLevel = authPriv
+                self.privProtocol = privProtocol
+                self.privKey = authProtocol.computeKey(privSecret or secret)
+
+            self.authProtocol = authProtocol
+            self.authKey = authProtocol.computeKey(authSecret or secret)
+
+class LocalizedCredentials:
     def __init__(self,
         auth: Optional[AuthProtocol] = None,
         priv: Optional[PrivProtocol] = None,
@@ -185,235 +226,6 @@ class Credentials:
         self.auth = auth
         self.priv = priv
 
-class UserTable:
-    def __init__(self) -> None:
-        self.engines: Dict[bytes, Dict[bytes, Credentials]] = {}
-        self.lock = threading.Lock()
-
-    def addUser(self,
-        engineID: bytes,
-        userName: bytes,
-        credentials: Credentials,
-    ) -> None:
-        with self.lock:
-            try:
-                users = self.engines[engineID]
-            except KeyError:
-                users = dict()
-                self.engines[engineID] = users
-
-            users[userName] = credentials
-
-    def getUser(self, engineID: bytes, userName: bytes) -> Credentials:
-        with self.lock:
-            try:
-                users = self.engines[engineID]
-            except KeyError as err:
-                raise InvalidEngineID(engineID) from err
-
-            try:
-                return users[userName]
-            except KeyError as err:
-                raise InvalidUserName(userName) from err
-
-class UserBasedSecurityModule(SecurityModule):
-    MODEL = SecurityModel.USM
-
-    def __init__(self, engineID: Optional[bytes] = None) -> None:
-        self.engineID = engineID
-        self.timekeeper = TimeKeeper()
-        self.users = UserTable()
-
-        if self.engineID is not None:
-            self.timekeeper.update(self.engineID)
-
-    def addUser(self,
-        engineID: bytes,
-        userName: bytes,
-        auth: Optional[AuthProtocol] = None,
-        priv: Optional[PrivProtocol] = None,
-    ) -> None:
-        if auth is None and priv is not None:
-            priv = None
-
-        self.users.addUser(engineID, userName, Credentials(auth, priv))
-
-    def prepareOutgoing(self,
-        header: bytes,
-        data: bytes,
-        engineID: bytes,
-        securityName: bytes,
-        securityLevel: SecurityLevel,
-    ) -> bytes:
-        if securityLevel.auth:
-            user = self.users.getUser(engineID, securityName)
-
-            if not user.auth:
-                userName = securityName.decode
-                errmsg = f"Authentication is disabled for user \"{userName}\""
-                raise InvalidSecurityLevel(errmsg)
-
-            engineTimeParameters = self.timekeeper.getEngineTime(engineID)
-            snmpEngineBoots, snmpEngineTime = engineTimeParameters
-            msgAuthenticationParameters = user.auth.msgAuthenticationParameters
-            msgPrivacyParameters = b''
-
-            if securityLevel.priv:
-                if not user.priv:
-                    userName = securityName.decode
-                    errmsg = f"Privacy is disabled for user \"{userName}\""
-                    raise InvalidSecurityLevel(errmsg)
-
-                msgPrivacyParameters, ciphertext = user.priv.encrypt(
-                    data,
-                    snmpEngineBoots,
-                    snmpEngineTime,
-                )
-
-                data = OctetString(ciphertext).encode()
-
-        else:
-            if engineID == self.engineID:
-                engineTimeParameters = self.timekeeper.getEngineTime(engineID)
-                snmpEngineBoots, snmpEngineTime = engineTimeParameters
-            else:
-                snmpEngineBoots = 0
-                snmpEngineTime = 0
-
-            msgAuthenticationParameters = b''
-            msgPrivacyParameters = b''
-
-        encodedPrivacyParams = OctetString(msgPrivacyParameters).encode()
-        securityParameters = encode(
-            SEQUENCE,
-            b''.join((
-                OctetString(engineID).encode(),
-                Integer(snmpEngineBoots).encode(),
-                Integer(snmpEngineTime).encode(),
-                OctetString(securityName).encode(),
-                OctetString(msgAuthenticationParameters).encode(),
-                encodedPrivacyParams,
-            ))
-        )
-
-        msgSecurityParameters = OctetString(securityParameters).encode()
-        body = b''.join((header, msgSecurityParameters, data))
-        wholeMsg = encode(SEQUENCE, body)
-
-        if securityLevel.auth:
-            signature = cast(AuthProtocol, user.auth).sign(wholeMsg)
-            endIndex = len(wholeMsg) - len(data) - len(encodedPrivacyParams)
-            startIndex = endIndex - len(signature)
-            wholeMsg = b''.join((
-                wholeMsg[:startIndex],
-                signature,
-                wholeMsg[endIndex:]
-            ))
-
-        return wholeMsg
-
-    def processIncoming(self,
-        msg: subbytes,
-        securityLevel: SecurityLevel,
-        timestamp: Optional[float] = None,
-    ) -> Tuple[SecurityParameters, bytes]:
-        if timestamp is None:
-            timestamp = time()
-
-        msgSecurityParameters, msgData = \
-            OctetString.decode(msg, leftovers=True, copy=False)
-
-        ptr = decode(
-            msgSecurityParameters.data,
-            expected=SEQUENCE,
-            leftovers=False,
-            copy=False,
-        )
-
-        msgAuthoritativeEngineID, ptr = OctetString.decode(ptr, leftovers=True)
-        msgAuthoritativeEngineBoots, ptr  = Integer.decode(ptr, leftovers=True)
-        msgAuthoritativeEngineTime,  ptr  = Integer.decode(ptr, leftovers=True)
-        msgUserName,              ptr = OctetString.decode(ptr, leftovers=True)
-
-        msgAuthenticationParameters, ptr = \
-            OctetString.decode(ptr, leftovers=True)
-        msgAuthenticationParametersIndex = \
-            ptr.start - len(msgAuthenticationParameters.data)
-        msgPrivacyParameters = OctetString.decode(ptr)
-
-        engineID = cast(bytes, msgAuthoritativeEngineID.data)
-        userName = cast(bytes, msgUserName.data)
-        securityParameters = SecurityParameters(engineID, userName)
-
-        if not securityLevel.auth:
-            self.timekeeper.update(
-                engineID,
-                msgAuthoritativeEngineBoots.value,
-                msgAuthoritativeEngineTime.value,
-                timestamp=timestamp
-            )
-
-            return securityParameters, msgData[:]
-
-        try:
-            user = self.users.getUser(engineID, userName)
-        except InvalidEngineID as err:
-            raise UnknownEngineID(engineID) from err
-        except InvalidUserName as err:
-            raise UnknownUserName(userName) from err
-
-        if user.auth is None:
-            username = userName.decode()
-            errmsg = f"Authentication is disabled for user \"{username}\""
-            raise UnsupportedSecLevel(errmsg)
-        elif securityLevel.priv and user.priv is None:
-            username = userName.decode()
-            errmsg = f"Data privacy is disabled for user \"{username}\""
-            raise UnsupportedSecLevel(errmsg)
-
-        padding = user.auth.msgAuthenticationParameters
-        if len(msgAuthenticationParameters.data) != len(padding):
-            raise WrongDigest("Invalid signature length")
-
-        wholeMsg = b''.join((
-            msg.data[:msgAuthenticationParametersIndex],
-            padding,
-            msg.data[msgAuthenticationParametersIndex + len(padding):]
-        ))
-
-        if user.auth.sign(wholeMsg) != msgAuthenticationParameters.data:
-            raise WrongDigest("Invalid signature")
-
-        try:
-            if not self.timekeeper.updateAndVerify(
-                engineID,
-                msgAuthoritativeEngineBoots.value,
-                msgAuthoritativeEngineTime.value,
-                timestamp=timestamp,
-            ):
-                raise NotInTimeWindow((
-                    engineID,
-                    msgAuthoritativeEngineBoots.value,
-                    msgAuthoritativeEngineTime.value,
-                ))
-        except InvalidEngineID as err:
-            raise UnknownEngineID(engineID) from err
-
-        if securityLevel.priv:
-            try:
-                payload = cast(PrivProtocol, user.priv).decrypt(
-                    cast(bytes, OctetString.decode(msgData).data),
-                    msgAuthoritativeEngineBoots.value,
-                    msgAuthoritativeEngineTime.value,
-                    cast(bytes, msgPrivacyParameters.data),
-                )
-            except ValueError as err:
-                raise DecryptionError(str(err)) from err
-        else:
-            payload = msgData[:]
-
-        return securityParameters, payload
-
 class DiscoveredEngine:
     def __init__(self) -> None:
         self.namespace: Optional[str] = None
@@ -442,10 +254,41 @@ class DiscoveredEngine:
         self.refCount -= 1
         return self.refCount == 0
 
+class UserTable:
+    def __init__(self) -> None:
+        self.engines: Dict[bytes, Dict[bytes, LocalizedCredentials]] = {}
+
+    def assignCredentials(self,
+        engineID: bytes,
+        userName: bytes,
+        credentials: LocalizedCredentials,
+    ) -> None:
+        try:
+            users = self.engines[engineID]
+        except KeyError:
+            users = dict()
+            self.engines[engineID] = users
+
+        users[userName] = credentials
+
+    def getCredentials(self,
+        engineID: bytes,
+        userName: bytes,
+    ) -> LocalizedCredentials:
+        try:
+            users = self.engines[engineID]
+        except KeyError as err:
+            raise InvalidEngineID(engineID) from err
+
+        try:
+            return users[userName]
+        except KeyError as err:
+            raise InvalidUserName(userName) from err
+
 class UserEntry:
     def __init__(self,
         defaultSecurityLevel: SecurityLevel,
-        credentials: Mapping[str, Any],
+        credentials: Credentials
     ) -> None:
         self.credentials = credentials
         self.defaultSecurityLevel = defaultSecurityLevel
@@ -467,36 +310,147 @@ class NameSpace:
     def __setitem__(self, key: str, item: UserEntry) -> None:
         return self.users.__setitem__(key, item)
 
-class UsmAdmin:
-    def __init__(self) -> None:
-        self.lock = threading.Lock()
-        self.engines: Dict[bytes, DiscoveredEngine] = {}
-        self.namespaces: Dict[str, NameSpace] = {}
+class UsmSecurityParameters(Sequence):
+    def __init__(self,
+        engineID: bytes,
+        engineBoots: int,
+        engineTime: int,
+        userName: bytes,
+        signature: Asn1Data,
+        salt: bytes,
+    ):
+        self.engineID = engineID
+        self.engineBoots = engineBoots
+        self.engineTime = engineTime
+        self.userName = userName
+        self.salt = salt
 
-        self.securityModule = UserBasedSecurityModule()
+        if isinstance(signature, subbytes):
+            self.signature = signature[:]
+            self.signatureIndex = signature.start
+            self.wholeMsg = signature.data
+        else:
+            self.signature = signature
+            self.signatureIndex = None
+            self.wholeMsg = None
+
+    def __iter__(self) -> Iterator[Asn1Encodable]:
+        yield OctetString(self.engineID)
+        yield Integer(self.engineBoots)
+        yield Integer(self.engineTime)
+        yield OctetString(self.userName)
+        yield OctetString(self.signature)
+        yield OctetString(self.salt)
+
+    def __len__(self) -> int:
+        return 6
+
+    def __repr__(self) -> str:
+        args = (
+            str(self.engineID),
+            str(self.engineBoots),
+            str(self.engineTime),
+            str(self.userName),
+            str(self.signature),
+            str(self.salt),
+        )
+
+        return f"{typename(self)}({', '.join(args)})"
+
+    def __str__(self) -> str:
+        return self.toString()
+
+    def toString(self, depth: int = 0, tab: str = "    ") -> str:
+        indent = tab * depth
+        subindent = indent + tab
+        return "\n".join((
+            f"{indent}{typename(self)}:",
+            f"{subindent}Authoritative Engine ID: {self.engineID!r}",
+            f"{subindent}Authoritative Engine Boots: {self.engineBoots}",
+            f"{subindent}Authoritative Engine Time: {self.engineTime}",
+            f"{subindent}User Name: {self.userName}",
+            f"{subindent}Signature: {self.signature}",
+            f"{subindent}Encryption Salt: {self.salt}",
+        ))
+
+    @classmethod
+    def decode(cls, data, leftovers=False, copy=False, **kwargs):
+        return super().decode(data, leftovers, copy, **kwargs)
+
+    @classmethod
+    def deserialize(cls, data: Asn1Data):
+        copy = not isinstance(data, subbytes)
+
+        engineID, ptr = OctetString.decode(data, leftovers=True)
+        engineBoots, ptr   = Integer.decode(ptr, leftovers=True)
+        engineTime,  ptr   = Integer.decode(ptr, leftovers=True)
+        userName,  ptr = OctetString.decode(ptr, leftovers=True)
+        signature, ptr = OctetString.decode(ptr, leftovers=True, copy=copy)
+        salt           = OctetString.decode(ptr)
+
+        return cls(
+            cast(bytes, engineID.data),
+            engineBoots.value,
+            engineTime.value,
+            cast(bytes, userName.data),
+            signature.data,
+            cast(bytes, salt.data),
+        )
+
+    @classmethod
+    def findSignature(self, msgSecurityParameters: subbytes) -> subbytes:
+        ptr = decode(msgSecurityParameters, expected=SEQUENCE, copy=False)
+        _, ptr = decode(ptr, expected=OCTET_STRING, leftovers=True, copy=False)
+        _, ptr = decode(ptr, expected=INTEGER,      leftovers=True, copy=False)
+        _, ptr = decode(ptr, expected=INTEGER,      leftovers=True, copy=False)
+        _, ptr = decode(ptr, expected=OCTET_STRING, leftovers=True, copy=False)
+        ptr, _ = decode(ptr, expected=OCTET_STRING, leftovers=True, copy=False)
+        return ptr
+
+class UserBasedSecurityModule(SecurityModule):
+    MODEL = SecurityModel.USM
+
+    def __init__(self, engineID: Optional[bytes] = None) -> None:
+        self.engineID = engineID
+        self.engines: Dict[bytes, DiscoveredEngine] = {}
+        self.lock = threading.Lock()
+        self.namespaces: Dict[str, NameSpace] = {}
+        self.timekeeper = TimeKeeper()
+        self.users = UserTable()
+
+        if self.engineID is not None:
+            self.timekeeper.update(self.engineID)
 
     @staticmethod
-    def localize(
+    def localizeCredentials(
         engineID: bytes,
-        authProtocol: Optional[Type[AuthProtocol]] = None,
-        authSecret: Optional[bytes] = None,
-        privProtocol: Optional[Type[PrivProtocol]] = None,
-        privSecret: Optional[bytes] = None,
-    ) -> Tuple[Optional[AuthProtocol], Optional[PrivProtocol]]:
+        credentials: Optional[Credentials] = None,
+    ) -> LocalizedCredentials:
+        if credentials is None:
+           credentials = Credentials()
+
         auth = None
         priv = None
 
-        if authProtocol is not None:
-            assert authSecret is not None
-            authKey = authProtocol.localize(authSecret, engineID)
-            auth = authProtocol(authKey)
+        if credentials.authProtocol is not None:
+            assert credentials.authKey is not None
+            auth = credentials.authProtocol(
+                credentials.authProtocol.localizeKey(
+                    credentials.authKey,
+                    engineID,
+                )
+            )
 
-            if privProtocol is not None:
-                assert privSecret is not None
-                privKey = authProtocol.localize(privSecret, engineID)
-                priv = privProtocol(privKey)
+            if credentials.privProtocol is not None:
+                assert credentials.privKey is not None
+                priv = credentials.privProtocol(
+                    credentials.authProtocol.localizeKey(
+                        credentials.privKey,
+                        engineID,
+                    )
+                )
 
-        return auth, priv
+        return LocalizedCredentials(auth, priv)
 
     def addUser(self,
         userName: str,
@@ -509,25 +463,24 @@ class UsmAdmin:
         defaultSecurityLevel: Optional[SecurityLevel] = None,
         namespace: str = "",
     ) -> None:
-        credentials: Dict[str, Any] = dict()
-        if authProtocol is None:
-            maxSecurityLevel = noAuthNoPriv
-        else:
-            if privProtocol is None:
-                maxSecurityLevel = authNoPriv
-            else:
-                maxSecurityLevel = authPriv
-                credentials["privProtocol"] = privProtocol
-                credentials["privSecret"] = privSecret or secret
-
-            credentials["authProtocol"] = authProtocol
-            credentials["authSecret"] = authSecret or secret
+        credentials = Credentials(
+            authProtocol,
+            authSecret,
+            privProtocol,
+            privSecret,
+            secret,
+        )
 
         if defaultSecurityLevel is None:
-            defaultSecurityLevel = maxSecurityLevel
-        elif defaultSecurityLevel > maxSecurityLevel:
+            defaultSecurityLevel = credentials.maxSecurityLevel
+        elif defaultSecurityLevel > credentials.maxSecurityLevel:
             errmsg = "Unable to support {} without the \"{}\" argument"
-            param = "privProtocol" if maxSecurityLevel.auth else "authProtocol"
+
+            if credentials.maxSecurityLevel.auth:
+                param = "privProtocol"
+            else:
+                param = "authProtocol"
+
             raise ValueError(errmsg.format(defaultSecurityLevel, param))
 
         with self.lock:
@@ -582,7 +535,10 @@ class UsmAdmin:
 
             raise ValueError(errmsg) from err
 
-    def registerRemoteEngine(self, engineID: bytes, namespace: str) -> bool:
+    def registerRemoteEngine(self,
+        engineID: bytes,
+        namespace: str = "",
+    ) -> bool:
         with self.lock:
             try:
                 engine = self.engines[engineID]
@@ -596,17 +552,18 @@ class UsmAdmin:
             if not initialized and assigned:
                 ns = self.namespaces[namespace]
                 for userName, entry in ns:
-                    auth, priv = self.localize(engineID, **entry.credentials)
-                    self.securityModule.addUser(
+                    self.users.assignCredentials(
                         engineID,
                         userName.encode(),
-                        auth,
-                        priv,
+                        self.localizeCredentials(engineID, entry.credentials)
                     )
 
             return assigned
 
-    def unregisterRemoteEngine(self, engineID: bytes, namespace: str) -> None:
+    def unregisterRemoteEngine(self,
+        engineID: bytes,
+        namespace: str = "",
+    ) -> None:
         with self.lock:
             try:
                 engine = self.engines[engineID]
@@ -615,3 +572,166 @@ class UsmAdmin:
             else:
                 if engine.release(namespace):
                     del self.engines[engineID]
+
+    def prepareOutgoing(self,
+        message: SNMPv3Message,
+        engineID: bytes,
+        securityName: bytes,
+        timestamp: Optional[float] = None,
+    ) -> bytes:
+        if timestamp is None:
+            timestamp = time()
+
+        if message.header.flags.authFlag:
+            with self.lock:
+                user = self.users.getCredentials(engineID, securityName)
+
+            if not user.auth:
+                userName = securityName.decode
+                errmsg = f"Authentication is disabled for user \"{userName}\""
+                raise InvalidSecurityLevel(errmsg)
+
+            snmpEngineBoots, snmpEngineTime = self.timekeeper.getEngineTime(
+                engineID,
+                timestamp=timestamp,
+            )
+
+            msgAuthenticationParameters = user.auth.msgAuthenticationParameters
+            msgPrivacyParameters = b''
+
+            if message.header.flags.privFlag:
+                if not user.priv:
+                    userName = securityName.decode
+                    errmsg = f"Privacy is disabled for user \"{userName}\""
+                    raise InvalidSecurityLevel(errmsg)
+
+                ciphertext, msgPrivacyParameters = user.priv.encrypt(
+                    message.scopedPDU.encode(),
+                    snmpEngineBoots,
+                    snmpEngineTime,
+                )
+
+                message.encryptedPDU = OctetString(ciphertext)
+
+        else:
+            if engineID == self.engineID:
+                engineTimeParameters = self.timekeeper.getEngineTime(
+                    engineID,
+                    timestamp=timestamp,
+                )
+
+                snmpEngineBoots, snmpEngineTime = engineTimeParameters
+            else:
+                snmpEngineBoots = 0
+                snmpEngineTime = 0
+
+            msgAuthenticationParameters = b''
+            msgPrivacyParameters = b''
+
+        securityParameters = UsmSecurityParameters(
+            engineID,
+            snmpEngineBoots,
+            snmpEngineTime,
+            securityName,
+            msgAuthenticationParameters,
+            msgPrivacyParameters,
+        )
+
+        message.securityParameters = OctetString(securityParameters.encode())
+        wholeMsg = message.encode()
+
+        if message.header.flags.authFlag:
+            location = UsmSecurityParameters.findSignature(
+                SNMPv3Message.findSecurityParameters(wholeMsg)
+            )
+
+            signature = cast(AuthProtocol, user.auth).sign(wholeMsg)
+            wholeMsg = location.replace(signature)
+
+        return wholeMsg
+
+    def processIncoming(self,
+        message: SNMPv3Message,
+        timestamp: Optional[float] = None,
+    ) -> None:
+        if timestamp is None:
+            timestamp = time()
+
+        securityParameters = UsmSecurityParameters.decode(
+            message.securityParameters.data,
+        )
+
+        message.securityEngineID = securityParameters.engineID
+        message.securityName     = securityParameters.userName
+
+        remoteIsAuthoritative = (securityParameters.engineID != self.engineID)
+
+        if not message.header.flags.authFlag:
+            if remoteIsAuthoritative:
+                self.timekeeper.update(
+                    securityParameters.engineID,
+                    securityParameters.engineBoots,
+                    securityParameters.engineTime,
+                    timestamp=timestamp
+                )
+
+            return
+
+        try:
+            with self.lock:
+                user = self.users.getCredentials(
+                    securityParameters.engineID,
+                    securityParameters.userName,
+                )
+        except InvalidEngineID as err:
+            raise UnknownEngineID(securityParameters.engineID) from err
+        except InvalidUserName as err:
+            raise UnknownUserName(securityParameters.userName) from err
+
+        if user.auth is None:
+            username = securityParameters.userName.decode()
+            errmsg = f"Authentication is disabled for user \"{username}\""
+            raise UnsupportedSecLevel(errmsg)
+        elif message.header.flags.privFlag and user.priv is None:
+            username = securityParameters.userName.decode()
+            errmsg = f"Data privacy is disabled for user \"{username}\""
+            raise UnsupportedSecLevel(errmsg)
+
+        padding = user.auth.msgAuthenticationParameters
+        if len(securityParameters.signature) != len(padding):
+            raise WrongDigest("Invalid signature length")
+
+        wholeMsg = bytearray(securityParameters.wholeMsg)
+
+        start = securityParameters.signatureIndex
+        stop = start + len(padding)
+        wholeMsg[start:stop] = padding
+
+        if user.auth.sign(wholeMsg) != securityParameters.signature:
+            raise WrongDigest("Invalid signature")
+
+        try:
+            if not self.timekeeper.updateAndVerify(
+                securityParameters.engineID,
+                securityParameters.engineBoots,
+                securityParameters.engineTime,
+                timestamp=timestamp,
+            ):
+                raise NotInTimeWindow((
+                    securityParameters.engineID,
+                    securityParameters.engineBoots,
+                    securityParameters.engineTime,
+                ))
+        except InvalidEngineID as err:
+            raise UnknownEngineID(securityParameters.engineID) from err
+
+        if message.header.flags.privFlag:
+            try:
+                message.plaintext = cast(PrivProtocol, user.priv).decrypt(
+                    cast(bytes, message.encryptedPDU.data),
+                    securityParameters.engineBoots,
+                    securityParameters.engineTime,
+                    securityParameters.salt,
+                )
+            except ValueError as err:
+                raise DecryptionError(str(err)) from err
