@@ -1,13 +1,20 @@
-__all__ = ["TransportDomain", "TransportListener"]
+__all__ = ["TransportDomain", "TransportListener", "TransportMultiplexor"]
 
 from abc import abstractmethod
 from collections import namedtuple
 import enum
 import os
+import socket
 
 from snmp.typing import *
 
-TransportDomain = enum.Enum("TransportDomain", ("UDP",))
+class TransportDomain(enum.Enum):
+    def __init__(self, family: socket.AddressFamily, loopback: str):
+        self.address_family = family
+        self.loopback_address = loopback
+
+    UDP_IPv4 = socket.AF_INET, "127.0.0.1"
+    UDP_IPv6 = socket.AF_INET6, "::1"
 
 T = TypeVar("T")
 class TransportLocator(Generic[T]):
@@ -32,20 +39,30 @@ class Transport(Generic[T]):
         ...
 
     @abstractmethod
-    def listen(self, listener: "TransportListener") -> None:
+    def send(self, address: T, data: bytes) -> None:
+        ...
+
+class TransportListener:
+    @abstractmethod
+    def hear(self, transport: Transport[T], address: T, data: bytes) -> None:
+        ...
+
+TransportType = TypeVar("TransportType", bound=Transport[Tuple[str, int]])
+class TransportMultiplexor(Generic[TransportType]):
+    @abstractmethod
+    def register(self, sock: TransportType) -> None:
         ...
 
     @abstractmethod
-    def send(self, address: T, data: bytes) -> None:
+    def listen(self, listener: TransportListener) -> None:
         ...
 
     @abstractmethod
     def stop(self) -> None:
         ...
 
-class TransportListener:
     @abstractmethod
-    def hear(self, transport: Transport[T], address: T, data: bytes) -> None:
+    def close(self) -> None:
         ...
 
 supported = ("posix")
